# -*- coding: utf-8 -*-

from argparse import Namespace
from collections import defaultdict
from pathlib import Path

import adr
from loguru import logger

from mozci.data.base import DataSource

here = Path(__file__).parent.resolve()
adr.sources.load_source(here)


class ActiveDataSource(DataSource):
    """Uses 'adr' to query ActiveData."""

    name = "adr"
    supported_contracts = (
        "push_tasks",
        "push_tasks_tags",
        "push_test_groups",
    )

    @classmethod
    def normalize(cls, result):
        if "header" in result:
            result["data"] = [
                {
                    field: value
                    for field, value in zip(result["header"], entry)
                    if value is not None
                }
                for entry in result["data"]
            ]

        items = []
        retries = defaultdict(int)

        for item in result["data"]:
            if "id" not in item:
                logger.trace(f"Skipping {item} because of missing id.")
                continue

            task_id = item["id"]

            # If a task is re-run, use the data from the last run.
            if "retry_id" in item:
                if item["retry_id"] < retries[task_id]:
                    logger.trace(f"Skipping {item} because there is a newer run of it.")
                    continue

                retries[task_id] = item["retry_id"]

                # We don't need to store the retry ID.
                del item["retry_id"]

            items.append(item)
        return items

    def run_push_tasks(self, **kwargs):
        result = adr.query.run_query("push_tasks_from_treeherder", Namespace(**kwargs))
        tasks = []

        # If we are missing one of these keys, discard the task.
        required_keys = (
            "id",
            "label",
            "result",
        )

        for task in self.normalize(result):
            missing = [k for k in required_keys if k not in task]
            taskstr = task.get("label", task["id"])

            if missing:
                logger.trace(
                    f"Skipping task '{taskstr}' because it is missing "
                    f"the following attributes: {', '.join(missing)}"
                )
                continue

            if task.get("tags"):
                task["tags"] = {t["name"]: t["value"] for t in task["tags"]}

            if task.get("classification_note"):
                if isinstance(task["classification_note"], list):
                    task["classification_note"] = task["classification_note"][-1]
                    if task["classification_note"] is None:
                        del task["classification_note"]

            tasks.append(task)
        return tasks

    def run_push_tasks_tags(self, **kwargs):
        result = adr.query.run_query("push_tasks_tags_from_task", Namespace(**kwargs))
        tags = defaultdict(dict)
        for item in self.normalize(result):
            if "tags" not in item:
                continue

            for tag in item["tags"]:
                if any(k not in tag for k in ("name", "value")):
                    continue
                tags[item["id"]][tag["name"]] = tag["value"]
        return tags

    def run_push_test_groups(self, **kwargs):
        result = adr.query.run_query(
            "push_test_groups_from_unittest", Namespace(**kwargs)
        )
        required_keys = ("result_group", "result_ok")
        groups = defaultdict(dict)
        for item in self.normalize(result):
            if any(k not in item for k in required_keys):
                continue

            groups[item["id"]][item["result_group"]] = bool(item["result_ok"])

        return groups



# -*- coding: utf-8 -*-
from functools import lru_cache

import requests
from requests.packages.urllib3.util.retry import Retry

# https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
DEFAULT_RETRIES = 5
DEFAULT_BACKOFF_FACTOR = 0.1
DEFAULT_STATUS_FORCELIST = [500, 502, 503, 504]


@lru_cache(maxsize=None)
def get_session(concurrency=50):
    session = requests.Session()

    retry = Retry(
        total=DEFAULT_RETRIES,
        backoff_factor=DEFAULT_BACKOFF_FACTOR,
        status_forcelist=DEFAULT_STATUS_FORCELIST,
    )

    # Default HTTPAdapter uses 10 connections. Mount custom adapter to increase
    # that limit. Connections are established as needed, so using a large value
    # should not negatively impact performance.
    http_adapter = requests.adapters.HTTPAdapter(
        pool_connections=concurrency, pool_maxsize=concurrency, max_retries=retry
    )
    session.mount("https://", http_adapter)
    session.mount("http://", http_adapter)

    return session



# -*- coding: utf-8 -*-

from abc import ABC, abstractproperty
from typing import Any, Dict, Tuple

from mozci.data.contract import all_contracts


class DataSource(ABC):
    def __init__(self) -> None:
        missing = [
            f"run_{c}"
            for c in self.supported_contracts
            if not hasattr(self, f"run_{c}")
        ]
        if missing:
            missing_str = "  \n".join(missing)
            raise Exception(
                f"{self.__class__.__name__} must define the following methods:\n{missing_str}"
            )

    @abstractproperty
    def name(self) -> str:
        pass

    @abstractproperty
    def supported_contracts(self) -> Tuple[str, ...]:
        pass

    def get(self, name: str, **kwargs: Any) -> Dict[Any, Any]:
        fn = getattr(self, f"run_{name}")
        return fn(**kwargs)


class DataHandler:
    ALL_SOURCES: Dict[str, DataSource] = {}

    def __init__(self, *sources: str) -> None:
        self.sources = [self.ALL_SOURCES[sname] for sname in sources]

    def get(self, name: str, **context: Any) -> Dict[Any, Any]:
        """Given a contract, find the first registered source that supports it
        run it and return the results.

        Args:
            name (str): Name of the contract to run.
            context (dict): Context to pass into the contract as defined by `Contract.schema_in`.

        Returns:
            dict: The output of the contract as defined by `Contract.schema_out`.
        """
        if name not in all_contracts:
            raise Exception(f"Contract {name} does not exist!")

        # Validate input.
        contract = all_contracts[name]
        contract.validate_in(context)

        source = None
        for src in self.sources:
            if name in src.supported_contracts:
                source = src
                break
        else:
            raise Exception(f"No registered sources support {name}!")

        result = source.get(name, **context)

        # Validate output.
        contract.validate_out(result)
        return result


def register_sources():
    from mozci.data.sources import activedata, treeherder

    DataHandler.ALL_SOURCES = {
        "treeherder": treeherder.TreeherderSource(),
        "adr": activedata.ActiveDataSource(),
    }
